
from typing import cast
from typing import List

from logging import Logger
from logging import getLogger

# noinspection PyPackageRequirements
from deprecated import deprecated

from wx import MouseEvent
from wx import Font
from wx import FONTFAMILY_SWISS
from wx import FONTSTYLE_NORMAL
from wx import FONTWEIGHT_NORMAL

from miniogl.RectangleShape import RectangleShape
from miniogl.ShapeEventHandler import ShapeEventHandler

from ogl.OglLink import OglLink
from ogl.OglUtils import OglUtils

from ogl.events.OglEvents import OglEventType
from ogl.events.OglEventEngine import OglEventEngine

from ogl.preferences.OglPreferences import OglPreferences

from pyutmodel.PyutObject import PyutObject


DEFAULT_FONT_SIZE = 10


class OglObject(RectangleShape, ShapeEventHandler):
    """
    This is the base class for new OGL objects.
    Every new OGL class must inherit this class and redefine methods if
    necessary. OGL Objects are automatically a RectangleShape for
    global link management.
    """

    clsLogger: Logger = getLogger(__name__)

    def __init__(self, pyutObject=None, width: int = 0, height: int = 0):
        """

        Args:
            pyutObject: Associated PyutObject
            width:      Initial width
            height:     Initial height
        """
        self._pyutObject = pyutObject
        RectangleShape.__init__(self, 0, 0, width, height)

        # Default font
        self._defaultFont: Font           = Font(DEFAULT_FONT_SIZE, FONTFAMILY_SWISS, FONTSTYLE_NORMAL, FONTWEIGHT_NORMAL)
        self._prefs:       OglPreferences = OglPreferences()

        # TODO This is also used by sequence diagrams to store OglSDMessage links
        self._oglLinks: List[OglLink] = []     # Connected links
        self._modifyCommand = None

        self._eventEngine: OglEventEngine = cast(OglEventEngine, None)

    @deprecated(reason='Use the properties')
    def setPyutObject(self, pyutObject: PyutObject):
        self._pyutObject = pyutObject

    @deprecated(reason='Use the properties')
    def getPyutObject(self) -> PyutObject:
        """
        Return the associated pyut object.

        @return PyutObject : Associated PyutObject
        """
        return self._pyutObject

    @property
    def pyutObject(self):
        return self._pyutObject

    @pyutObject.setter
    def pyutObject(self, pyutObject):
        self._pyutObject = pyutObject

    @property
    def links(self):
        return self._oglLinks

    @property
    def eventEngine(self) -> OglEventEngine:
        """
        This property necessary because the diagram is not added until the
        object is' attached'

        Returns:

        """
        if self.HasDiagramFrame() is True:

            # from org.pyut.ui.UmlDiagramsFrame import UmlDiagramsFrame
            # panel: UmlDiagramsFrame = self.GetDiagram().GetPanel()
            panel = self.GetDiagram().GetPanel()
            if panel is not None:
                if self._eventEngine is None:
                    self._eventEngine = panel.eventEngine

        return self._eventEngine

    def addLink(self, link):
        """
        Add a link to an ogl object.

        Args:
            link:  the link to add
        """
        self._oglLinks.append(link)

    @deprecated(reason='Use the property links')
    def getLinks(self):
        """
        Return the links.

        Returns: OglLink[] : Links connected to object
        """
        return self._oglLinks

    def OnLeftDown(self, event: MouseEvent):
        """
        Handle event on left click.
        Note to self.  This method used to call only call event.Skip() if there was an action waiting
        Now I do it regardless;  Seem to be no ill effects

        Args:
            event:  The mouse event
        """
        OglObject.clsLogger.debug(f'OglObject.OnLeftDown  - {event.GetEventObject()=}')

        self.eventEngine.sendEvent(OglEventType.ShapeSelected, selectedShape=self, selectedShapePosition=event.GetPosition())
        event.Skip()

    def OnLeftUp(self, event: MouseEvent):
        """
        Implement this method so we can snap Ogl objects

        Args:
            event:  the mouse event
        """
        gridInterval: int = self._prefs.backgroundGridInterval
        x, y = self.GetPosition()
        if self._prefs.snapToGrid is True:
            snappedX, snappedY = OglUtils.snapCoordinatesToGrid(x=x, y=y, gridInterval=gridInterval)
            self.SetPosition(snappedX, snappedY)

    def autoResize(self):
        """
        Find the right size to see all the content and resize self.

        """
        pass

    def SetPosition(self, x: int, y: int):
        """
        Define new position for the object

        Args:
            x:  The new abscissa
            y:  The new ordinate
        """
        if self.eventEngine is not None:        # we might be associated with a diagram yet
            self.eventEngine.sendEvent(OglEventType.DiagramFrameModified)
        RectangleShape.SetPosition(self, x, y)

    def SetSelected(self, state=True):

        # from org.pyut.ui.Mediator import Mediator          # avoid circular import
        # from org.pyut.ui.Mediator import ACTION_ZOOM_OUT   # avoid circular import

        # mediator: Mediator = Mediator()
        # if mediator.getCurrentAction() != ACTION_ZOOM_OUT:
        # TODO:  I took this out because could never cause this to happen;  Conveniently, this removes all mediator calls
        RectangleShape.SetSelected(self, state)




from logging import Logger
from logging import getLogger

from pyutmodel.PyutLinkType import PyutLinkType

from ogl.Singleton import Singleton
from ogl.OglAssociation import OglAssociation
from ogl.OglAggregation import OglAggregation
from ogl.OglComposition import OglComposition
from ogl.OglInheritance import OglInheritance
from ogl.OglInterface import OglInterface
from ogl.OglNoteLink import OglNoteLink

from ogl.sd.OglSDMessage import OglSDMessage


def getOglLinkFactory():
    """
    Function to get the unique OglLinkFactory instance (singleton).
    """
    return OglLinkFactory()


def getLinkType(link: OglAssociation) -> PyutLinkType:
    """

    Args:
        link:   The enumeration OglLinkType

    Returns:  The OglLinkType

    """

    if isinstance(link, OglAggregation):
        return PyutLinkType.AGGREGATION
    elif isinstance(link, OglComposition):
        return PyutLinkType.COMPOSITION
    elif isinstance(link, OglInheritance):
        return PyutLinkType.INHERITANCE
    elif isinstance(link, OglAssociation):
        return PyutLinkType.ASSOCIATION
    elif isinstance(link, OglInterface):
        return PyutLinkType.INTERFACE
    elif isinstance(link, OglNoteLink):
        return PyutLinkType.NOTELINK


class OglLinkFactory(Singleton):
    """
    This class is a factory to produce `OglLink` objects.
    It works under the Factory Design Pattern model. Ask for a link
    from this object, and it will return you what you was asking for.
    """
    def init(self, *args, **kwargs):
        # noinspection PyAttributeOutsideInit
        self.logger: Logger = getLogger(__name__)

    def getOglLink(self, srcShape, pyutLink, destShape, linkType: PyutLinkType):
        """
        Used to get a OglLink of the given linkType.

        Args:
            srcShape:   Source shape
            pyutLink:   Conceptual links associated with the graphical links.
            destShape:  Destination shape
            linkType:   The linkType of the link (OGL_INHERITANCE, ...)

        Returns:  The requested link
        """
        if linkType == PyutLinkType.AGGREGATION:
            return OglAggregation(srcShape, pyutLink, destShape)

        elif linkType == PyutLinkType.COMPOSITION:
            return OglComposition(srcShape, pyutLink, destShape)

        elif linkType == PyutLinkType.INHERITANCE:
            return OglInheritance(srcShape, pyutLink, destShape)

        elif linkType == PyutLinkType.ASSOCIATION:
            return OglAssociation(srcShape, pyutLink, destShape)

        elif linkType == PyutLinkType.INTERFACE:
            return OglInterface(srcShape, pyutLink, destShape)

        elif linkType == PyutLinkType.NOTELINK:
            return OglNoteLink(srcShape, pyutLink, destShape)

        elif linkType == PyutLinkType.SD_MESSAGE:
            return OglSDMessage(srcShape=srcShape, pyutSDMessage=pyutLink, dstShape=destShape)
        else:
            self.logger.error(f"Unknown OglLinkType: {linkType}")
            return None




from typing import Tuple

from miniogl.ShapeModel import ShapeModel


class RectangleShapeModel(ShapeModel):
    """
    This class is the model of a RectangleShape ('view' in an MVC pattern).
    """
    def __init__(self, viewShape=None):
        """
        Used when the model is created first without any view.
        We have to use AddShape() and UpdateModel from the shape before
        we can use the model.

        Set the coordinates to 0 and a empty list of associated shapes (views)
        """
        super().__init__(viewShape)

        self._width:  int = 0
        self._height: int = 0

    def GetSize(self) -> Tuple[int, int]:
        """

        Returns:
            the size of the model
        """
        return self._width, self._height

    def SetSize(self, width: int, height: int):
        """
        Set the size of the model

        Args:
            width:      width of the model
            height:     height of the model
        """
        self._width = width
        self._height = height



from types import MethodType


class Singleton(object):
    """
    Base class for singleton classes.

    Any class derived from this one is a singleton. You can call its
    constructor multiple times, you'll get only one instance.

    Note that `__init__` must not be defined. Use `init` instead.
    This is because `__init__` will always be called, thus reinitializing the
    state of your singleton each time you try to instantiate it.
    On the contrary, `init` will be called just one time.

    To be sure that the `__init__` method won't be inadvertently defined,
    the singleton will check for it at first instantiation and raise an
    `AssertionError` if it finds it.

    Example::

        # This class is OK
        class A(Singleton):
            def init(self, theNewValue):
                self.theNewValue = theNewValue

        # This class will raise AssertionError at first instantiation, because
        # of the __init__ method.
        class B(Singleton):
            def __init__(self, theNewValue):
                self.theNewValue = theNewValue
    """
    def __new__(cls, *args, **kwargs):
        """
        New operator of a singleton class.
        Will return the only instance, or create it if needed.
        """
        instance = cls.__dict__.get("__instance__")
        if instance is None:
            instance = object.__new__(cls)
            assert type(instance.__init__) != MethodType, f"Error, your singleton class {cls} cannot contain a __init__ method."  # type: ignore
            try:
                instance.init(*args, **kwargs)
            except Exception as e:
                raise e
            cls.__instance__ = instance
        return instance

    def init(self, *args, **kwargs):
        """
        Constructor of a singleton class.
        """
        pass



from typing import cast

from wx import Colour
from wx import DC
from wx import Pen

from miniogl.Shape import Shape


class PointShape(Shape):

    SELECTION_ZONE: int = 8     # Make it bigger than in legacy;  It was 5 then
    """
    A point, which is drawn as a little square (3 pixels wide).

    """
    def __init__(self, x: int, y: int, parent=None):
        """

        Args:
            x:  x position of the point
            y:  y position of the point
            parent:  parent shape
        """
        super().__init__(x, y, parent)
        self._selectZone:           int = PointShape.SELECTION_ZONE
        self._visibleWhenSelected: bool = True

        self.__penSaveColor: Colour = cast(Colour, None)

    def Draw(self, dc: DC, withChildren=True):
        """
        Draw the point on the dc.

        Args:
            dc:
            withChildren:
        """
        if self._visible or (self._visibleWhenSelected and self._selected):

            self.__penSaveColor = dc.GetPen().GetColour()
            Shape.Draw(self, dc, False)

            self.__resetPenColor(dc)

            x, y = self.GetPosition()
            if not self._selected:
                dc.DrawRectangle(x - 1, y - 1, 3, 3)
            else:
                dc.DrawRectangle(x - 3, y - 3, 7, 7)
            if withChildren:
                self.DrawChildren(dc)

    def GetSelectionZone(self) -> int:
        """
        Get the selection tolerance zone, in pixels.

        Returns: half of the selection zone.
        """

        return self._selectZone

    def SetSelectionZone(self, halfWidth: int):
        """
        Set the selection tolerance zone, in pixels.

        Args:
            halfWidth: half of the selection zone.
        """
        self._selectZone = halfWidth

    def Inside(self, x: int, y: int):
        """

        Args:
            x: x coordinate
            y: y coordinate

        Returns:          `True` if (x, y) is inside the shape.

        """
        ax, ay = self.GetPosition()     # GetPosition always returns absolute position
        zone = self._selectZone
        return (ax - zone < x < ax + zone) and (ay - zone < y < ay + zone)

    def SetVisibleWhenSelected(self, state: bool):
        """
        Set to True if you want the point to always be visible when it's selected.

        @param  state
        """
        self._visibleWhenSelected = state

    def GetVisibleWhenSelected(self):
        """
        Return the "visible when selected flag".

        @return bool True if the shape is always visible when selected
        """
        return self._visibleWhenSelected

    def __resetPenColor(self, dc: DC):

        pen: Pen = dc.GetPen()
        pen.SetColour(self.__penSaveColor)
        dc.SetPen(pen)




from typing import List

from miniogl.PointShape import PointShape


class LinePoint(PointShape):
    """
    This is a point guiding a line.
    """
    def __init__(self, x: int, y: int, parent=None):
        """

        Args:
            x:  abscissa of point
            y:  ordinate of point
            parent:     parent shape
        """
        #  print ">>>LinePoint ", x, y
        super().__init__(x, y, parent)
        self._lines: List = []    # a list of LineShape(s) passing through this point

    def AddLine(self, line):
        """
        Add a line to this point.

        @param  line
        """
        self._lines.append(line)

    def Detach(self):
        """
        Detach the point from the diagram
        This also removes the point from all the lines it belongs to.
        """
        PointShape.Detach(self)
        for line in self._lines:
            line.Remove(self)
        self._lines = []

    def GetLines(self):
        """
        Get the lines passing through this point.
        Modifying the returned list won't modify the point itself.

        @return LineShape []
        """
        return self._lines[:]

    def RemoveLine(self, line):
        """
        Remove a line from this point.

        @param line
        """
        if line in self._lines:
            self._lines.remove(line)

    def SetMoving(self, state: bool):
        """
        A non-moving shape will be redrawn faster when others are moved.
        See DiagramFrame.Refresh for more information.

        @param  state
        """
        PointShape.SetMoving(self, state)
        for line in self._lines:
            line.SetMoving(state)




from typing import List
from typing import cast
from typing import Tuple

from logging import Logger
from logging import getLogger

from wx import BLACK_PEN
from wx import Brush
from wx import DC
from wx import Font
from wx import Pen
from wx import RED
from wx import RED_PEN
from wx import WHITE_BRUSH

from deprecated import deprecated

from miniogl.ShapeModel import ShapeModel

from ogl.preferences.OglPreferences import OglPreferences


class Shape:
    """
    Shape is the basic graphical block. It is also the view in
    an MVC pattern, so it has a relative model (ShapeModel).
    """

    ID = 0  # internal ID number
    clsLogger: Logger = getLogger(__name__)

    def __init__(self, x: int = 0, y: int = 0, parent=None):
        """
        If a parent is given, the position is relative to the parent's origin.

        Args:
            x: position of the shape on the diagram
            y: position of the shape on the diagram
            parent:
        """
        self._x: int = x    # shape position (view)
        self._y: int = y    # shape position (view)
        self._ox: int = 0   # origin position (view)
        self._oy: int = 0   # origin position (view)

        self._parent    = parent     # parent shape
        self._selected:  bool = False      # is the shape selected ?
        self._visible:   bool = True       # is the shape visible ?
        self._draggable: bool = True       # can the shape be dragged ?
        self._moving:    bool = False      # is this shape moving now ?
        self._protected: bool = False      # to protect against deletion

        self._anchors:         List = []   # anchors of the shape
        self._children:        List = []   # child shapes
        self._privateChildren: List = []   # private children, not saved

        self._pen:   Pen   = BLACK_PEN    # pen to use
        self._brush: Brush = WHITE_BRUSH  # brush to use

        self._model: ShapeModel = ShapeModel(self)  # model of the shape (MVC pattern)

        from miniogl.Diagram import Diagram

        self._diagram: Diagram = cast(Diagram, None)       # associated diagram

        self._id = Shape.ID     # unique ID number
        Shape.ID += 1
        if OglPreferences().debugBasicShape is True:
            from miniogl.TextShape import TextShape
            from miniogl.LineShape import LineShape
            if isinstance(self, (TextShape, LineShape)) is False:
                t: TextShape = self.AddText(0, -10, str(self._id))
                t.SetColor(RED)

    @property
    def id(self) -> int:
        return self._id

    @id.setter
    def id(self, newValue: int):
        self._id = newValue

    def SetPen(self, pen: Pen):
        """
        Set the pen used to draw the shape.

        @param pen
        """
        self._pen = pen

    def GetPen(self):
        """
        Get the pen used to draw the shape.

        @return wx.Pen
        """
        return self._pen

    def SetBrush(self, brush: Brush):
        """
        Set the brush used to draw the shape.

        @param brush
        """
        self._brush = brush

    def GetBrush(self):
        """
        Get the brush used to draw the shape.

        @return wx.Brush
        """
        return self._brush

    @deprecated(reason='Use the .id property')
    def GetID(self):

        return self._id

    @deprecated(reason='Use the .id property')
    def SetID(self, theId):
        """
        TODO: Make this a property

        Set the ID of the shape. This can be used when loading a shape.

        @param theId
        """
        self._id = theId
        Shape.ID = max(self._id + 1, Shape.ID)

    def SetOrigin(self, x: int, y: int):
        """
        Set the origin of the shape, from its upper left corner.

        Args:
            x:  new origin
            y:  new origin

        """
        self._ox, self._oy = x, y

    def GetOrigin(self):
        """
        Get the origin of the shape, from its upper left corner.

        @return double x, y : origin
        """
        return self._ox, self._oy

    def AppendChild(self, child: "Shape"):
        """
        Append a child to this shape.
        This doesn't add it to the diagram, but it will be drawn by the parent.

        @param  child
        """
        child.SetParent(self)
        self._children.append(child)

    def GetAllChildren(self):
        """
        Get all the children of this shape, recursively.

        @return Shape []
        """
        shapes = []
        for child in self._children:
            shapes.append(child.GetAllChildren())
        return shapes

    def GetChildren(self):
        """
        Get the children of this shape.
        Not recursively, only get first level children.
        It's a copy of the original list, modifying it won't modify the
        original.

        @return Shape []
        """
        return self._children[:]

    def AddAnchor(self, x: int, y: int, anchorType=None):
        """
        Add an anchor point to the shape.
        A line can be linked to it. The anchor point will stay bound to the
        shape and move with it. It is protected against deletion (by default)
        and not movable by itself.

        Args:
            x: position of the new point, relative to the origin of the shape
            y: position of the new point, relative to the origin of the shape
            anchorType:  class to use as anchor point, or None for default (AnchorPoint)

        Returns:    the created anchor
        """
        from miniogl.AnchorPoint import AnchorPoint     # I don't like in module imports but there is a cyclical dependency somewhere

        if anchorType is None:
            anchorType = AnchorPoint
        p = anchorType(x, y, self)
        p.SetProtected(True)
        self._anchors.append(p)
        if self._diagram is not None:
            self._diagram.AddShape(p)
        # if the shape is not yet attached to a diagram, the anchor points
        # will be attached when Attach is called on the shape.
        return p

    def AddAnchorPoint(self, anchor):
        """
        Add an anchor point directly.

        Args:
            anchor:
        """
        self._anchors.append(anchor)

    def RemoveAllAnchors(self):
        """
        Remove all anchors of the shape.
        """
        while self._anchors:
            self.RemoveAnchor(self._anchors[0])

    def RemoveAnchor(self, anchor):
        """
        Remove an anchor.

        Args:
            anchor:   The anchor to remove
        """
        if anchor in self._anchors:
            self._anchors.remove(anchor)

    def AddText(self, x: int, y: int, text: str, font: Font = None):
        """
        Add a text shape to the shape.
        Args:
            x : position of the text, relative to the origin of the shape
            y : position of the text, relative to the origin of the shape
            text : text to add
            font: font to use

        Returns:
            The created shape
        """
        # Shape.clsLogger.debug(f'AddText - {text=} @ ({x}, {y})')
        t = self._createTextShape(x, y, text, font=font)
        self._children.append(t)

        return t

    def Attach(self, diagram):
        """
        Don't use this method, use Diagram.AddShape instead !!!
        Attach the shape to a diagram.
        When you create a new shape, you must attach it to a diagram before
        you can see it. This method is used internally by Diagram.AddShape.

        @param  diagram
        """
        self._diagram = diagram
        # add the anchors and the children
        map(lambda x: diagram.AddShape(x), self._anchors + self._children + self._privateChildren)

    def Detach(self):
        """
        Detach the shape from its diagram.
        This is the way to delete a shape. All anchor points are also
        removed, and link lines too.
        """
        # do not detach a protected shape
        from miniogl.Diagram import Diagram

        if self._diagram is not None and not self._protected:
            # noinspection PyProtectedMember
            self.GetModel()._views.remove(self)

            diagram = self._diagram
            self._diagram = cast(Diagram, None)
            diagram.RemoveShape(self)
            # detach the anchors + children
            while self._anchors:
                child = self._anchors[0]
                child.SetProtected(False)
                child.Detach()
                child.SetProtected(True)
            for child in self._children + self._privateChildren:
                child.SetProtected(False)
                child.Detach()
                child.SetProtected(True)

            # Shape.clsLogger.debug("now, the shapes are", diagram.GetShapes())

    def Draw(self, dc: DC, withChildren: bool = True):
        """
        Draw the shape.
        For a shape, only the anchors are drawn. Nothing is drawn if the
        shape is set invisible.
        For children classes, the main classes would normally call its
        parent's Draw method, passing withChildren = False, and finally
        calling itself the DrawChildren method.

        Args:
            dc:             wxPython device context
            withChildren:   draw the children or not
        """

        if self._visible:
            dc.SetPen(self._pen)
            dc.SetBrush(self._brush)
            if withChildren:
                self.DrawChildren(dc)

        if self._selected:
            dc.SetPen(RED_PEN)
            self.DrawHandles(dc)

    def DrawChildren(self, dc: DC):
        """
        Draw the children of this shape.

        Args:
            dc:
        """
        if self._visible:
            for child in self._children + self._anchors + self._privateChildren:
                # Shape.clsLogger.debug(f'DrawChildren {child=} {child._selected=}')
                child.Draw(dc)

    def DrawBorder(self, dc):
        """
        Draw the border of the shape, for fast rendering.

        @param dc
        """
        pass

    def DrawAnchors(self, dc):
        """
        Draw the anchors of the shape.

        @param dc
        """
        #  print "Shape.DrawAnchors; shape=", self, ", anchors=", self._anchors
        map(lambda x: x.Draw(dc), self._anchors)

    def DrawHandles(self, dc: DC):
        """
        Draw the handles (selection points) of the shape.
        A shape has no handles, because it has no size.

        @param  dc
        """
        pass

    def GetAnchors(self):
        """
        Return a list of the anchors of the shape.

        @return AnchorPoint []
        """
        return self._anchors[:]     # Funky Python notation that creates a copy

    def GetParent(self):
        """
        Get the parent of this shape.

        @return Shape parent
        """
        return self._parent

    def SetParent(self, parent: 'Shape'):
        """
        Set the parent of this shape.

        @param parent
        """
        self._parent = parent

    def GetPosition(self) -> Tuple[int, int]:
        """
        Return the absolute position of the shape.
        It is in the diagram's coordinate system.

        Returns: An x,y tuple

        """
        if self._parent is not None:
            x, y = self._parent.GetPosition()
            return self._x + x, self._y + y
        else:
            return self._x, self._y

    def GetTopLeft(self):
        """
        Get the coordinates of the top left point in diagram coordinates.

        @return (double, double)
        """
        x, y = self.GetPosition()
        x -= self._ox
        y -= self._oy
        return x, y

    def GetSize(self) -> Tuple[int, int]:
        """
        Get the size of the shape.

        Returns:  a width, height tuple

        """
        return 0, 0

    def ConvertCoordToRelative(self, x, y):
        """
        Convert absolute coordinates to relative ones.
        Relative coordinates are coordinates relative to the origin of the
        shape.

        @return (double, double)
        """
        if self._parent is not None:
            ox, oy = self._parent.GetPosition()
            x -= ox
            y -= oy
        return x, y

    def GetRelativePosition(self):
        """
        Return the position of the shape, relative to its parent.

        @return (double, double)
        """
        return self._x, self._y

    def Inside(self, x, y) -> bool:
        """

        Args:
            x: x coordinate
            y: y coordinate

        Returns:          `True` if (x, y) is inside the shape.
        """
        return False

    def IsDraggable(self):
        """
        True if the shape can be dragged.

        @return bool
        """
        return self._draggable

    def IsProtected(self):
        """
        True if the shape is protected.

        @return bool
        """
        return self._protected

    def IsSelected(self):
        """
        True if the shape is selected.

        @return bool
        """
        return self._selected

    def IsVisible(self):
        """
        True if the shape is visible.

        @return bool
        """
        return self._visible

    def SetSelected(self, state: bool = True):
        """
        Set the shape selected state

        Args:
            state: `True` if it is selected else `False`
        """
        self._selected = state

    def IsMoving(self):
        """
        Return the "moving" state of a shape.
        See SetMoving.
        """
        return self._moving

    def SetMoving(self, state):
        """
        A non-moving shape will be redrawn faster when others are moved.
        See DiagramFrame.Refresh for more information.

        @param state
        """
        self._moving = state
        for shape in self._children:
            shape.SetMoving(state)
        for anchor in self._anchors:
            anchor.SetMoving(state)

    def SetDraggable(self, drag):
        """
        If False, the shape won't be movable.

        @param drag
        """
        self._draggable = drag

    def SetPosition(self, x: int, y: int):
        """
        If it's draggable; change the position of the shape;  Upper left corner

        Args:
            x:  x position to move shape to
            y:  y position to move shape to
        """
        if self._draggable:
            if self._parent is None:
                self._x = x
                self._y = y
                # Shape.clsLogger.debug(f'{self._id=} Position: ({self._x},{self._y})')
            else:
                # Shape.clsLogger.debug(f'_parent: {self._parent}')
                self._x, self._y = self.ConvertCoordToRelative(x, y)
            #  if the shape is attached to a diagramFrame, it means that
            #  the model will be initialized correctly.
            # (Avoid a null pointer error).
            if self.HasDiagramFrame():
                self.UpdateModel()

    def SetRelativePosition(self, x: int, y: int):
        """
        Set the position of the shape, relative to the parent.
        Only works if the shape is draggable.

        Args:
            x:
            y:
        """
        if self._draggable:
            self._x = x
            self._y = y

    def SetProtected(self, newValue: bool):
        """
        Protect the shape against deletion (Detach).

        @param newValue
        """
        self._protected = newValue

    def SetSize(self, w: int, h: int):
        """
        Set the size of the shape.

        Args:
            w: width
            h: height of the shape
        """
        pass

    def SetVisible(self, theNewValue: bool):
        """
        Set the shape visible or not.

        @param theNewValue
        """
        self._visible = theNewValue

    def GetDiagram(self):
        """
        Return the diagram associated with this shape.

        @return Diagram
        """
        return self._diagram

    def UpdateFromModel(self):
        """
        Updates the shape position from the model in the light of a
        change of state of the diagram frame (here it's only for the zoom)
        """

        # Get the coordinates of the model (ShapeModel)
        mx, my = self.GetModel().GetPosition()

        # Get the offsets and the ratio between the shape (view) and the
        # shape model (ShapeModel) given by the frame where the shape
        # is displayed.
        ratio = self.GetDiagram().GetPanel().currentZoom
        dx: int = round(self.GetDiagram().GetPanel().xOffSet)
        dy: int = round(self.GetDiagram().GetPanel().yOffSet)

        # calculation of the shape (view) coordinates in the light of the offsets and ratio
        x: int = round(ratio * mx) + dx
        y: int = round(ratio * my) + dy

        # assign the new coordinates to the shape (view). DON'T USE SetPosition(),
        # because there is a call to UpdateModel() in that method.
        if self._parent is not None:
            self._x, self._y = self.ConvertCoordToRelative(x, y)
        else:
            self._x = x
            self._y = y

    def UpdateModel(self):
        """
        Updates the coordinates of the model (ShapeModel) when the Shape (view)
        is moved.
        """

        #  get the associated model (ShapeModel)
        model = self.GetModel()

        # Get the offsets and the ratio between the shape (view) and the
        # shape model (ShapeModel) given by the frame where the shape
        # is displayed.
        from miniogl.DiagramFrame import DiagramFrame
        diagram = self.GetDiagram()
        panel: DiagramFrame   = diagram.GetPanel()   # to enable debugging and unit tests

        # ratio = self.GetDiagram().GetPanel().GetCurrentZoom()
        # dx = self.GetDiagram().GetPanel().GetXOffset()
        # dy = self.GetDiagram().GetPanel().GetYOffset()
        ratio = panel.currentZoom
        dx    = panel.xOffSet
        dy    = panel.yOffSet

        #  get the coordinates of this shape
        x, y = self.GetPosition()

        # calculation of the model coordinates in the light of the
        # offsets and ratio and assignment.
        mx = round((x - dx) // ratio)
        my = round((y - dy) // ratio)
        model.SetPosition(mx, my)

        # change also the position of the model of the children,
        # because when we move the parent children set position is not called
        # and so their update model is not called
        for child in self._anchors:
            cx, cy = child.GetPosition()
            cmx = round((cx - dx) // ratio)
            cmy = round((cy - dy) // ratio)
            child.GetModel().SetPosition(cmx, cmy)

    def GetModel(self):
        """

        Returns:  he model of the shape (MVC pattern)
        """
        return self._model

    def SetModel(self, modelShape):
        """
        Set the specified model associated to this shape (MVC pattern)

        Args:
            modelShape:  model to set
        """
        self._model = modelShape

    def HasDiagramFrame(self):
        """

        Returns: `True` if the shape has a diagram and if this diagram has
        a diagram frame.
        """
        if self.GetDiagram() is not None:
            return self.GetDiagram().GetPanel() is not None
        else:
            return False

    def _addPrivateText(self, x: int, y: int, text: str, font: Font = None):
        """
        Add a text shape, putting it in the private children of the shape.
        It won't be saved !!! This is used in constructor of child classes.

        Args:
            x,: position of the text, relative to the origin of the shape
            y : position of the text, relative to the origin of the shape
            text: text to add
            font: font to use

        Returns:  TextShape : the created shape
        """
        from miniogl.TextShape import TextShape

        t: TextShape = self._createTextShape(x, y, text, font=font)
        self._privateChildren.append(t)
        return t

    def _createTextShape(self, x: int, y: int, text: str, font: Font = None):
        """
        Create a text shape and add it to the diagram.

        Args:
            x,: position of the text, relative to the origin of the shape
            y : position of the text, relative to the origin of the shape
            text: text to add
            font: font to use

        Returns:  TextShape : the created shape

        """
        from miniogl.TextShape import TextShape

        textShape: TextShape = TextShape(x, y, text, self, font=font)
        if self._diagram is not None:
            self._diagram.AddShape(textShape)
        # Shape.clsLogger.debug(f'_CreateTextShape - {textShape=} added at {x=} {y=}')

        return textShape




from enum import Enum


class OglTextFontFamily(Enum):
    """
    How we specify fonts for Pyut
    """
    SWISS    = 'Swiss'
    MODERN   = 'Modern'
    ROMAN    = 'Roman'
    SCRIPT   = 'Script'
    TELETYPE = 'Teletype'



# type: ignore
"""
This is a suite of small classes used to draw RotatableShapes.
Each one represent a simple shape (line, rectangle, circle, ellipse...) or
abstract command (color change).
"""


class VShape:
    """
    Base VShape class.

    """
    def __init__(self):
        self._data = ()

    def Convert(self, angle, x, y):
        T = [
            [1, 0, 0, 1], [0, -1, 1, 0], [-1, 0, 0, -1], [0, 1, -1, 0]
        ]
        nx = T[angle][0] * x + T[angle][1] * y
        ny = T[angle][2] * x + T[angle][3] * y
        return nx, ny

    def SetAngle(self, angle):
        pass

    def Scale(self, scale, data):
        return map(lambda x: x * scale, data)


class VRectangle(VShape):
    def __init__(self, x, y, w, h):
        VShape.__init__(self)
        self._data = (x, y, w, h)

    def SetAngle(self, angle):
        x, y, w, h = self._data
        x, y = self.Convert(angle, x, y)
        w, h = self.Convert(angle, w, h)
        self._data = (x, y, w, h)

    def Draw(self, dc, ox, oy, scale):
        if scale == 1:
            x, y, w, h = self._data
        else:
            x, y, w, h = self.Scale(scale, self._data)
        dc.DrawRectangle(ox + x, oy + y, w, h)


class VEllipse(VShape):
    def __init__(self, x, y, w, h):
        VShape.__init__(self)
        self._data = (x, y, w, h)

    def SetAngle(self, angle):
        x, y, w, h = self._data
        x, y = self.Convert(angle, x, y)
        w, h = self.Convert(angle, w, h)
        self._data = (x, y, w, h)

    def Draw(self, dc, ox, oy, scale):
        if scale == 1:
            x, y, w, h = self._data
        else:
            x, y, w, h = self.Scale(scale, self._data)
        dc.DrawEllipse(ox + x, oy + y, w, h)


class VCircle(VShape):
    def __init__(self, x, y, r):
        VShape.__init__(self)
        self._data = (x, y, r)

    def SetAngle(self, angle):
        x, y, r = self._data
        x, y = self.Convert(angle, x, y)
        self._data = (x, y, r)

    def Draw(self, dc, ox, oy, scale):
        if scale == 1:
            x, y, r = self._data
        else:
            x, y, r = self.Scale(scale, self._data)
        dc.DrawCircle(ox + x, oy + y, r)


class VArc(VShape):
    def __init__(self, x1, y1, x2, y2, xc, yc):
        VShape.__init__(self)
        self._data = (x1, y1, x2, y2, xc, yc)

    def SetAngle(self, angle):
        x1, y1, x2, y2, xc, yc = self._data
        x1, y1 = self.Convert(angle, x1, y1)
        x2, y2 = self.Convert(angle, x2, y2)
        xc, yc = self.Convert(angle, xc, yc)
        self._data = (x1, y1, x2, y2, xc, yc)

    def Draw(self, dc, ox, oy, scale):
        if scale == 1:
            x1, y1, x2, y2, xc, yc = self._data
        else:
            x1, y1, x2, y2, xc, yc = self.Scale(scale, self._data)
        dc.DrawArc(ox + x1, oy + y1, ox + x2, oy + y2, ox + xc, oy + yc)


class VEllipticArc(VShape):
    def __init__(self, x, y, w, h, start, end):
        VShape.__init__(self)
        self._data = (x, y, w, h, start, end)

    def SetAngle(self, angle):
        x, y, w, h, start, end  = self._data
        x, y = self.Convert(angle, x, y)
        w, h = self.Convert(angle, w, h)
        start -= angle * 90
        end -= angle * 90
        self._data = (x, y, w, h, start, end)

    def Draw(self, dc, ox, oy, scale):
        if scale == 1:
            x, y, w, h, start, end  = self._data
        else:
            x, y, w, h = self.Scale(scale, self._data[0:4])
            start, end = self._data[4:]
        dc.DrawEllipticArc(ox + x, oy + y, w, h, start, end)


class VLineLength(VShape):
    def __init__(self, x, y, w, h):
        VShape.__init__(self)
        self._data = (x, y, w, h)

    def SetAngle(self, angle):
        x, y, w, h = self._data
        x, y = self.Convert(angle, x, y)
        w, h = self.Convert(angle, w, h)
        self._data = (x, y, w, h)

    def Draw(self, dc, ox, oy, scale):
        if scale == 1:
            x, y, w, h = self._data
        else:
            x, y, w, h = self.Scale(scale, self._data)
        x, y = ox + x, oy + y
        dc.DrawLine(x, y, x + w, y + h)


class VLineDest(VShape):
    def __init__(self, sx, sy, dx, dy):
        VShape.__init__(self)
        self._data = (sx, sy, dx, dy)

    def SetAngle(self, angle):
        sx, sy, dx, dy = self._data
        sx, sy = self.Convert(angle, sx, sy)
        dx, dy = self.Convert(angle, dx, dy)
        self._data = (sx, sy, dx, dy)

    def Draw(self, dc, ox, oy, scale):
        if scale == 1:
            sx, sy, dx, dy = self._data
        else:
            sx, sy, dx, dy = self.Scale(scale, self._data)
        dc.DrawLine(ox + sx, oy + sy, ox + dx, oy + dy)


class VPolygon(VShape):
    def __init__(self, points):
        VShape.__init__(self)
        self._data = points

    def SetAngle(self, angle):
        new = []
        for x, y in self._data:
            x, y = self.Convert(angle, x, y)
            new.append((x, y))
        self._data = tuple(new)

    def Draw(self, dc, ox, oy, scale):
        if scale == 1:
            points = self._data
        else:
            points = []
            for x, y in self._data:
                points.append(tuple(self.Scale(scale, (x, y))))
        dc.DrawPolygon(points, ox, oy)


class VPen(VShape):
    def __init__(self, pen):
        VShape.__init__(self)
        self._pen = pen

    # noinspection PyUnusedLocal
    def Draw(self, dc, x, y, scale=1):
        dc.SetPen(self._pen)


class VBrush(VShape):
    def __init__(self, brush):
        VShape.__init__(self)
        self._brush = brush

    # noinspection PyUnusedLocal
    def Draw(self, dc, x, y, scale=1):
        dc.SetBrush(self._brush)
